@@ -1,23 +1,24 @@
 import numpy as np
-from ase import Atoms
 from itertools import combinations
 import os
 import logging
-from typing import List, Dict, Tuple, Any, Union
+from typing import List, Dict, Tuple, Any, Union, Optional
+from xyzgraph import DATA
 
 from . import config
+from .utils import calculate_distance, calculate_angle, calculate_dihedral
 
 logger = logging.getLogger("vib_analysis")
 
-def read_xyz_trajectory(file_path: str) -> List[Atoms]:
+def read_xyz_trajectory(file_path: str) -> List[Dict[str, Any]]:
     """
-    Read an XYZ trajectory file and return a list of ASE Atoms objects.
+    Read an XYZ trajectory file and return a list of frame dicts.
     
     Args:
         file_path: Path to XYZ trajectory file
         
     Returns:
-        List of ASE Atoms objects, one per frame
+        List of frame dicts with 'symbols' and 'positions' keys, one per frame
         
     Raises:
         FileNotFoundError: If file doesn't exist
@@ -41,7 +42,10 @@ def read_xyz_trajectory(file_path: str) -> List[Atoms]:
                 parts = f.readline().split()
                 symbols.append(parts[0])
                 coords.append([float(x) for x in parts[1:]])
-            frame = Atoms(symbols=symbols, positions=coords)
+            frame = {
+                'symbols': symbols,
+                'positions': np.array(coords)
+            }
             frames.append(frame)
     
     if len(frames) == 1:
@@ -55,33 +59,48 @@ def read_xyz_trajectory(file_path: str) -> List[Atoms]:
     return frames
 
 def build_internal_coordinates(
-    frame: Atoms, 
+    frame: Dict[str, Any],
+    displaced_frames: Optional[List[Dict[str, Any]]] = None,
+    independent_graphs: bool = False,
+    ig_flexible: bool = config.IG_FLEXIBLE_DEFAULT,
+    relaxed: bool = config.RELAXED,
     bond_tolerance: float = config.BOND_TOLERANCE,
-) -> Dict[str, List[Tuple[int, ...]]]:
+) -> Dict[str, Any]:
     """
     Build internal coordinates (bonds, angles, dihedrals) using xyzgraph with hierarchical thresholds.
-    
-    Uses three separate graphs with progressively tighter thresholds:
+    Uses two separate graphs:
     - Bond graph (flexible): captures forming/breaking bonds in TS
     - Tighter graph: more conservative connectivity for angles/dihedrals
+
+    When independent_graphs=True:
+    - TS connectivity built with bond_tolerance (flexible)
+    - Displaced connectivity built with xyzgraph defaults (strict) or bond_tolerance (if ig_flexible=True)
+    - All connectivity merged via union
+    - Bond coordinates derived from merged connectivity
+    - Angles/dihedrals derived from tighter TS connectivity for consistency
     
     Args:
-        frame: ASE Atoms object
+        frame: Frame dict with 'symbols' and 'positions' keys
+        displaced_frames: Optional list of displaced frame dicts for connectivity augmentation
+        independent_graphs: If True, augment TS connectivity with displaced connectivity
+        ig_flexible: If True (with independent_graphs), apply bond_tolerance to displaced graphs
+        relaxed: Use relaxed rules for xyzgraph (for complex ring systems)
         bond_tolerance: Multiplier for bond detection (most flexible)
         
     Returns:
-        Dictionary with keys 'bonds', 'angles', 'dihedrals' containing lists of
-        atom index tuples
+        Dictionary with keys 'bonds', 'angles', 'dihedrals', 'connectivity'
     """
     from xyzgraph import build_graph
     
-    # Convert ASE Atoms to xyzgraph format
-    atoms_list = [(frame.get_chemical_symbols()[i], tuple(frame.positions[i])) 
-                  for i in range(len(frame))]
+    # Convert frame dict to xyzgraph format
+    symbols = frame['symbols']
+    positions = frame['positions']
+    atoms_list = [(symbols[i], tuple(positions[i])) 
+                  for i in range(len(symbols))]
 
     # xyzgraph base thresholds for flexible bond graph
     threshold_h_nm = 0.42 * bond_tolerance
-    threshold_nm_nm = 0.55 * bond_tolerance 
+    threshold_nm_nm = 0.5 * bond_tolerance 
     threshold_h_m = 0.45 * bond_tolerance 
     threshold_m_l = 0.65 # Keep metal-ligand as is - no scaling
     threshold_h_h = 0.40 * bond_tolerance 
@@ -95,16 +114,71 @@ def build_internal_coordinates(
         threshold_h_metal=threshold_h_m,
         threshold_metal_ligand=threshold_m_l,
         threshold_h_h=threshold_h_h,
+        relaxed=relaxed,  # Use relaxed rules if specified
         quick=True,  # Fast mode - we don't need bond orders
         method='cheminf',
         debug=False
     )
     
     # Extract bonds from flexible graph
-    bonds = [(i, j) for i, j in G_bonds.edges()]
-    logger.debug(f"Built bond graph with {len(bonds)} bonds (tolerance={bond_tolerance})")
+    ts_bonds = set(G_bonds.edges())
+    logger.debug(f"Built TS bond graph with {len(ts_bonds)} bonds (tolerance={bond_tolerance})")
     
-    # 2. Build tighter graph for angles/dihedrals with default xyzgraph thresholds
+    # 2. If independent_graphs, augment with displaced connectivity
+    if independent_graphs and displaced_frames:
+        logger.debug(f"Augmenting TS connectivity with displaced graphs (flexible={ig_flexible})")
+        
+        augmented_bonds = set(ts_bonds)  # Start with TS
+        
+        # Determine thresholds for displaced graphs
+        if ig_flexible:
+            # Use bond_tolerance (same as TS)
+            disp_h_nm = 0.42 * bond_tolerance
+            disp_nm_nm = 0.5 * bond_tolerance
+            disp_h_m = 0.45 * bond_tolerance
+            disp_h_h = 0.40 * bond_tolerance
+            logger.debug("Using flexible thresholds for displaced graphs")
+        else:
+            # Use xyzgraph defaults
+            disp_h_nm = 0.42
+            disp_nm_nm = 0.5
+            disp_h_m = 0.45
+            disp_h_h = 0.40
+            logger.debug("Using default thresholds for displaced graphs")
+        
+        # Build displaced graphs and collect bonds
+        for idx, disp_frame in enumerate(displaced_frames):
+            disp_atoms = [(disp_frame['symbols'][i], tuple(disp_frame['positions'][i])) 
+                         for i in range(len(disp_frame['symbols']))]
+            
+            G_disp = build_graph(
+                atoms=disp_atoms,
+                threshold=1.0,
+                threshold_h_nonmetal=disp_h_nm,
+                threshold_nonmetal_nonmetal=disp_nm_nm,
+                threshold_h_metal=disp_h_m,
+                threshold_metal_ligand=threshold_m_l,
+                threshold_h_h=disp_h_h,
+                relaxed=relaxed,
+                quick=True,
+                method='cheminf',
+                debug=False
+            )
+            
+            disp_bonds = set(G_disp.edges())
+            new_bonds = disp_bonds - ts_bonds
+            if new_bonds:
+                logger.debug(f"Displaced frame {idx}: found {len(new_bonds)} new bonds")
+            augmented_bonds.update(disp_bonds)
+        
+        logger.debug(f"Augmented connectivity: {len(ts_bonds)} TS bonds + {len(augmented_bonds - ts_bonds)} new = {len(augmented_bonds)} total")
+        bonds = list(augmented_bonds)
+    else:
+        # Standard behavior: TS only
+        bonds = list(ts_bonds)
+    
+    # 3. Build tighter graph for angles/dihedrals with default xyzgraph thresholds
+    # Always use TS geometry for consistency
     G_tight = build_graph(
         atoms=atoms_list,
         threshold=1.0,  # Use xyzgraph defaults
@@ -146,21 +220,6 @@ def build_internal_coordinates(
         'connectivity': connectivity
     }
 
-def calculate_bond_length(frame: Atoms, i: int, j: int) -> float:
-    """Calculate bond length between atoms i and j."""
-    return round(float(frame.get_distance(i, j)), 3)
-
-
-def calculate_angle(frame: Atoms, i: int, j: int, k: int) -> float:
-    """Calculate angle between atoms i-j-k."""
-    return round(float(frame.get_angle(i, j, k, mic=True)), 3)
-
-
-def calculate_dihedral(frame: Atoms, i: int, j: int, k: int, l: int) -> float:
-    """Calculate dihedral angle for atoms i-j-k-l."""
-    return round(float(frame.get_dihedral(i, j, k, l, mic=True)), 3)
-
-
 def _has_significant_bond_change(
     bond: Tuple[int, int], 
     bond_changes: Dict[Tuple[int, int], Tuple[float, float]], 
@@ -205,14 +264,15 @@ def _bonds_are_stable(
 
 
 def calculate_internal_changes(
-    frames: List[Atoms],
-    ts_frame: Atoms,
-    internal_coords: Dict[str, List[Tuple[int, ...]]],
+    frames: List[Dict[str, Any]],
+    ts_frame: Dict[str, Any],
+    internal_coords: Dict[str, Any],
     bond_threshold: float = config.BOND_THRESHOLD,
     angle_threshold: float = config.ANGLE_THRESHOLD,
     dihedral_threshold: float = config.DIHEDRAL_THRESHOLD,
-    bond_stability_threshold: float = config.BOND_STABILITY_THRESHOLD
-) -> Tuple[Dict, Dict, Dict, Dict, Dict]:
+    coupled_motion_filter: float = config.COUPLED_MOTION_FILTER,
+    coupled_proton_threshold: Union[float, bool] = config.COUPLED_PROTON_THRESHOLD
+) -> Tuple[Dict, Dict, Dict, Dict, Dict, set]:
     """
     Track changes in internal coordinates across trajectory.
     
@@ -221,30 +281,68 @@ def calculate_internal_changes(
     to other structural changes.
     
     Args:
-        frames: List of trajectory frames (typically 2 most diverse)
-        ts_frame: Reference frame (typically transition state)
+        frames: List of frame dicts (typically 2 most diverse)
+        ts_frame: Reference frame dict (typically transition state)
         internal_coords: Dictionary with 'bonds', 'angles', 'dihedrals' lists
         bond_threshold: Minimum bond change to report (Å)
         angle_threshold: Minimum angle change to report (degrees)
         dihedral_threshold: Minimum dihedral change to report (degrees)
-        bond_stability_threshold: Threshold for filtering coupled angle/dihedral changes (Å)
+        coupled_motion_filter: Threshold for filtering coupled angle/dihedral changes (Å)
         
     Returns:
         Tuple of (bond_changes, angle_changes, minor_angles, 
-                  unique_dihedrals, dependent_dihedrals)
-        Each is a dict mapping coordinate tuple to (max_change, initial_value)
+                  unique_dihedrals, dependent_dihedrals, coupled_proton_bonds)
+        Each dict maps coordinate tuple to (max_change, initial_value)
+        coupled_proton_bonds is a set of bonds detected via coupled threshold
     """
+    # Get symbols from ts_frame for element identification
+    symbols = ts_frame['symbols']
+    
     # Identify significant bond changes
     bond_changes = {}
     for i, j in internal_coords['bonds']:
-        distances = [calculate_bond_length(frame, i, j) for frame in frames]
+        distances = [calculate_distance(frame['positions'], i, j) for frame in frames]
         max_change = round(max(distances) - min(distances), 3)
         if abs(max_change) >= bond_threshold:
-            initial_length = calculate_bond_length(ts_frame, i, j)
+            initial_length = calculate_distance(ts_frame['positions'], i, j)
             bond_changes[(i, j)] = (max_change, initial_length)
     
     logger.debug(f"Found {len(bond_changes)} significant bond changes")
     
+    # Coupled proton transfer detection (for any H involved in detected bond changes)
+    coupled_proton_bonds = set()  # Track which bonds were detected via coupled threshold
+    if coupled_proton_threshold is not False:
+        # Find ALL H atoms involved in any detected bond change
+        h_atoms_in_changes = set()
+        for (i, j) in bond_changes:
+            if symbols[i] == 'H':
+                h_atoms_in_changes.add(i)
+            if symbols[j] == 'H':
+                h_atoms_in_changes.add(j)
+        
+        if h_atoms_in_changes:
+            logger.debug(f"Searching for coupled proton transfers for {len(h_atoms_in_changes)} H atoms")
+        
+        # Search for additional bonds involving these H atoms with reduced threshold
+        for h_idx in h_atoms_in_changes:
+            for i, j in internal_coords['bonds']:
+                if h_idx not in (i, j):
+                    continue  # Skip bonds not involving this H
+                
+                sorted_bond = tuple(sorted((i, j)))
+                if sorted_bond in bond_changes:
+                    continue  # Already detected
+                
+                # Check with reduced threshold
+                distances = [calculate_distance(frame['positions'], i, j) for frame in frames]
+                max_change = round(max(distances) - min(distances), 3)
+                if abs(max_change) >= coupled_proton_threshold:
+                    initial_length = calculate_distance(ts_frame['positions'], i, j)
+                    bond_changes[sorted_bond] = (max_change, initial_length)  
+                    coupled_proton_bonds.add(sorted_bond)  # Track separately
+                    logger.debug(f"Coupled proton detection: bond {sorted_bond} (Δ={max_change} Å)")
+                    # break     # commented out to allow other checks of connectivity
+    
     # Track atoms involved in bond changes
     changed_atoms = set()
     for bond in bond_changes:
@@ -261,16 +359,16 @@ def calculate_internal_changes(
         if any(bond in bond_changes for bond in bonds_in_angle):
             continue
         
-        # Skip if constituent bonds are not stable
-        if not _bonds_are_stable(bonds_in_angle, bond_changes, bond_stability_threshold):
+        # Skip if bonds are not stable (some motion but not enough for reporting a bond change)
+        if not _bonds_are_stable(bonds_in_angle, bond_changes, coupled_motion_filter):
             continue
         
         # Calculate angle change
-        angles = [calculate_angle(frame, i, j, k) for frame in frames]
+        angles = [calculate_angle(frame['positions'], i, j, k) for frame in frames]
         max_change = round(max(angles) - min(angles), 3)
         
         if abs(max_change) >= angle_threshold:
-            initial_angle = calculate_angle(ts_frame, i, j, k)
+            initial_angle = calculate_angle(ts_frame['positions'], i, j, k)
             angle_atoms = set((i, j, k))
             
             # Classify as minor if involves atoms from bond changes
@@ -294,12 +392,12 @@ def calculate_internal_changes(
         if any(set(bond).issubset({i, j, k, l}) for bond in bond_changes):
             continue
         
-        # Skip if constituent bonds are not stable
-        if not _bonds_are_stable(bonds_in_dihedral, bond_changes, bond_stability_threshold):
+        # Skip if bonds are not stable (some motion but not enough for reporting a bond change)
+        if not _bonds_are_stable(bonds_in_dihedral, bond_changes, coupled_motion_filter):
             continue
         
         # Calculate dihedral change (adjust for periodicity)
-        dihedrals = [calculate_dihedral(frame, i, j, k, l) for frame in frames]
+        dihedrals = [calculate_dihedral(frame['positions'], i, j, k, l) for frame in frames]
         max_change = round(
             max([abs((d - dihedrals[0] + 180) % 360 - 180) for d in dihedrals]), 
             3
@@ -309,23 +407,25 @@ def calculate_internal_changes(
             dihedral_changes[(i, j, k, l)] = max_change
     
     # Group dihedrals by rotation axis and select representative
-    masses = frames[0].get_masses()
+    # Use atomic numbers as proxy for mass (heavier atoms have higher atomic numbers)
+    symbols = frames[0]['symbols']
+    atomic_numbers = [DATA.s2n[sym] for sym in symbols]
     dihedral_groups = {}
     
     for (i, j, k, l), change in dihedral_changes.items():
         axis = tuple(sorted((j, k)))
-        total_mass = masses[i] + masses[j] + masses[k] + masses[l]
+        total_atomic_number = atomic_numbers[i] + atomic_numbers[j] + atomic_numbers[k] + atomic_numbers[l]
         
         if axis not in dihedral_groups:
             dihedral_groups[axis] = []
-        dihedral_groups[axis].append(((i, j, k, l), change, total_mass))
+        dihedral_groups[axis].append(((i, j, k, l), change, total_atomic_number))
     
     # Select most significant dihedral per axis
     unique_dihedrals = {}
     dependent_dihedrals = {}
     
     for axis, dihedrals_list in dihedral_groups.items():
-        # Sort by mass and change magnitude
+        # Sort by atomic number sum and change magnitude
         dihedrals_sorted = sorted(
             dihedrals_list, 
             key=lambda x: (x[2], x[1]), 
@@ -334,7 +434,7 @@ def calculate_internal_changes(
         dihedral, max_change, _ = dihedrals_sorted[0]
         
         if max_change >= dihedral_threshold:
-            initial_dihedral = calculate_dihedral(ts_frame, *dihedral)
+            initial_dihedral = calculate_dihedral(ts_frame['positions'], *dihedral)
             dihedral_atoms = set(dihedral)
             
             # Classify as dependent if involves atoms from bond changes
@@ -348,15 +448,18 @@ def calculate_internal_changes(
         f"{len(dependent_dihedrals)} dependent dihedrals"
     )
     
-    return bond_changes, angle_changes, minor_angles, unique_dihedrals, dependent_dihedrals
+    if coupled_proton_bonds:
+        logger.debug(f"Found {len(coupled_proton_bonds)} bonds via coupled proton threshold")
+    
+    return bond_changes, angle_changes, minor_angles, unique_dihedrals, dependent_dihedrals, coupled_proton_bonds
 
-def compute_rmsd(frame1, frame2):
-    """Computes RMSD between two ASE frames."""
-    diff = frame1.get_positions() - frame2.get_positions()
+def compute_rmsd(frame1: Dict[str, Any], frame2: Dict[str, Any]) -> float:
+    """Computes RMSD between two frame dicts."""
+    diff = frame1['positions'] - frame2['positions']
     return np.sqrt(np.mean(np.sum(diff**2, axis=1)))
 
 
-def select_most_diverse_frames(frames, top_n=2):
+def select_most_diverse_frames(frames: List[Dict[str, Any]], top_n: int = 2) -> List[int]:
     """Select frames with largest RMSD from the TS frame (frame 0)."""
     # create an RMSD matrix between all frames and select the highest pair
     rmsd_matrix = np.zeros((len(frames), len(frames)))
@@ -379,13 +482,17 @@ def select_most_diverse_frames(frames, top_n=2):
     return selected_indices
 
 def analyze_internal_displacements(
-    xyz_file_or_frames: Union[str, List[Atoms]],
+    xyz_file_or_frames: Union[str, List[Dict[str, Any]]],
+    relaxed: bool = config.RELAXED,
     bond_tolerance: float = config.BOND_TOLERANCE,
     bond_threshold: float = config.BOND_THRESHOLD,
     angle_threshold: float = config.ANGLE_THRESHOLD,
     dihedral_threshold: float = config.DIHEDRAL_THRESHOLD,
-    bond_stability_threshold: float = config.BOND_STABILITY_THRESHOLD,
+    coupled_motion_filter: float = config.COUPLED_MOTION_FILTER,
+    coupled_proton_threshold: Union[float, bool] = config.COUPLED_PROTON_THRESHOLD,
     ts_frame: int = config.DEFAULT_TS_FRAME,
+    independent_graphs: bool = False,
+    ig_flexible: bool = config.IG_FLEXIBLE_DEFAULT,
 ) -> Dict[str, Any]:
     """
     Analyze vibrational displacements in trajectory to identify structural changes.
@@ -395,12 +502,12 @@ def analyze_internal_displacements(
     and tracks coordinate changes.
     
     Args:
-        xyz_file_or_frames: Path to XYZ trajectory file or list of ASE Atoms objects
+        xyz_file_or_frames: Path to XYZ trajectory file or list of frame dicts
         bond_tolerance: Multiplier for xyzgraph bond detection thresholds
         bond_threshold: Minimum bond change to report (Å)
         angle_threshold: Minimum angle change to report (degrees)
         dihedral_threshold: Minimum dihedral change to report (degrees)
-        bond_stability_threshold: Threshold for filtering coupled angle/dihedral changes (Å)
+        coupled_motion_filter: Threshold for filtering coupled angle/dihedral changes (Å)
         ts_frame: Index of transition state frame to use as reference
         
     Returns:
@@ -414,7 +521,7 @@ def analyze_internal_displacements(
             - atom_index_map: Dict mapping atom indices to symbols
             
     Raises:
-        TypeError: If xyz_file_or_frames is not str or list of Atoms
+        TypeError: If xyz_file_or_frames is not str or list of frame dicts
         FileNotFoundError: If file path doesn't exist
         ValueError: If trajectory has less than 2 frames
     """
@@ -426,31 +533,38 @@ def analyze_internal_displacements(
     else:
         raise TypeError(
             "xyz_file_or_frames must be a file path (str) or "
-            "list of ASE Atoms objects."
+            "list of frame dicts."
         )
 
-    internal_coords = build_internal_coordinates(
-        frame=frames[ts_frame],
-        bond_tolerance=bond_tolerance,
-    )
-    
+    # Select diverse frames FIRST (needed for independent_graphs mode)
     selected_indices = select_most_diverse_frames(frames)
     selected_frames = [frames[i] for i in selected_indices]
     
-    logger.info(f"Selected frames {selected_indices} for analysis")
+    logger.info(f"Using TS frame {ts_frame}, selected frames {selected_indices} for analysis")
+    
+    # Build internal coordinates with optional augmentation
+    internal_coords = build_internal_coordinates(
+        frame=frames[ts_frame],
+        displaced_frames=selected_frames if independent_graphs else None,
+        independent_graphs=independent_graphs,
+        ig_flexible=ig_flexible,
+        relaxed=relaxed,
+        bond_tolerance=bond_tolerance,
+    )
 
-    bond_changes, angle_changes, minor_angles, unique_dihedrals, dependent_dihedrals = calculate_internal_changes(
+    bond_changes, angle_changes, minor_angles, unique_dihedrals, dependent_dihedrals, coupled_proton_bonds = calculate_internal_changes(
         frames=selected_frames,
         ts_frame=frames[ts_frame],
         internal_coords=internal_coords,
         bond_threshold=bond_threshold,
         angle_threshold=angle_threshold,
         dihedral_threshold=dihedral_threshold,
-        bond_stability_threshold=bond_stability_threshold,
+        coupled_motion_filter=coupled_motion_filter,
+        coupled_proton_threshold=coupled_proton_threshold,
     )
 
     first_frame = frames[0]
-    symbols = first_frame.get_chemical_symbols()
+    symbols = first_frame['symbols']
     atom_index_map = {i: s for i, s in enumerate(symbols)}
 
     return {
@@ -462,4 +576,5 @@ def analyze_internal_displacements(
         "frame_indices": selected_indices,
         "atom_index_map": atom_index_map,
         "connectivity": internal_coords['connectivity'],
+        "coupled_proton_bonds": coupled_proton_bonds,
     }
